//
//  LevelEditor.m
//  CutTheVerlet
//
//  Created by Fahim Farook on 7/4/13.
//  Copyright (c) 2013 OpenTheJob. All rights reserved.
//

#import "LevelEditor.h"
#import "LevelFileHandler.h"
#import "PineappleModel.h"
#import "RopeModel.h"
#import "CutTheVerletGameLayer.h"
#import "RopeSprite.h"

enum {
    kEditMode,
    kRopeAnchorPineappleMode,
    kRopeAnchorAnyMode
} typedef editorMode;

@interface LevelEditor () {
    LevelFileHandler* fileHandler;
    CCSprite* background;
    CCSpriteBatchNode* pineapplesSpriteSheet;
    CCSpriteBatchNode* ropeSpriteSheet;
    NSMutableArray* ropes;
    UILongPressGestureRecognizer* longPressRecognizer;
    PopupMenu* popupMenu;
    editorMode mode;
    RopeModel* newRope;
    NSObject* selectedObject;
    CGPoint originalRopeAnchor;
    NSMutableArray* connectedRopes;
}

@end

@implementation LevelEditor

+(CCScene *)sceneWithFileHandler:(LevelFileHandler*)fileHandler {
    CCScene* scene = [CCScene node];
    LevelEditor *layer = [[LevelEditor alloc] initWithFileHandler:fileHandler];
	[scene addChild: layer];
    return scene;
}

-(id)initWithFileHandler:(LevelFileHandler*)levelFileHandler {
    self = [super init];
    if (self) {
        fileHandler = levelFileHandler;
        [self createMenu];
        
        background = [CCSprite spriteWithSpriteFrameName:@"bg.png"];
		CGSize winSize = [CCDirector sharedDirector].winSize;
        background.position = CGPointMake(winSize.width/2, winSize.height/2);
        background.color = kDefaultBackgroundColor;
		
        [self addChild:background];
        
        // Load the sprite sheet into the sprite cache
        [[CCSpriteFrameCache sharedSpriteFrameCache] addSpriteFramesWithFile:@"CutTheVerlet.plist"];
        pineapplesSpriteSheet = [CCSpriteBatchNode batchNodeWithFile:@"CutTheVerlet.pvr.ccz"];
        [self addChild:pineapplesSpriteSheet];
        ropeSpriteSheet = [CCSpriteBatchNode batchNodeWithFile:@"rope_texture.png"];
        [self addChild:ropeSpriteSheet];
        
        [self drawLoadedLevel];
    }
    return self;
}

-(void)onEnter {
    [super onEnter];
    [[CCDirector sharedDirector].touchDispatcher addTargetedDelegate:self priority:0 swallowsTouches:NO];
    longPressRecognizer = [[UILongPressGestureRecognizer alloc] initWithTarget:self action:@selector(longPress:)];
    UIView* openGLView = [CCDirector sharedDirector].view;
    [openGLView addGestureRecognizer: longPressRecognizer];
}

-(void)onExit {
    [super onExit];
    [[CCDirector sharedDirector].touchDispatcher removeDelegate:self];
    UIView* openGLView = [CCDirector sharedDirector].view;
    [openGLView removeGestureRecognizer: longPressRecognizer];
}

-(void)drawLoadedLevel {
    // Draw pineapple
    for (PineappleModel* pineapple in fileHandler.pineapples) {
        [self createPineappleSpriteFromModel:pineapple];
    }
    // Draw ropes
    ropes = [NSMutableArray arrayWithCapacity:5];
    for (RopeModel* ropeModel in fileHandler.ropes) {
        [self createRopeSpriteFromModel:ropeModel];
    }
}

-(void)createMenu {
    CGSize winSize = [CCDirector sharedDirector].winSize;
    
    // Place Buttons at bottom of game
    CCLabelTTF* saveLabel = [CCLabelTTF labelWithString:@"Save" fontName:@"Marker Felt" fontSize:24];
    CCMenuItem* saveItem = [CCMenuItemLabel itemWithLabel:saveLabel target:self selector:@selector(save)];
    
    CCLabelTTF* resetLabel = [CCLabelTTF labelWithString:@"Reset" fontName:@"Marker Felt" fontSize:24];
    CCMenuItem* resetItem = [CCMenuItemLabel itemWithLabel:resetLabel target:self selector:@selector(resetLevel)];
    
    CCLabelTTF* playLabel = [CCLabelTTF labelWithString:@"Play Level" fontName:@"Marker Felt" fontSize:24];
    CCMenuItem* playLevelItem = [CCMenuItemLabel itemWithLabel:playLabel target:self selector:@selector(playLevel)];
    
    // Create menu with buttons
    CCMenu* menu = [CCMenu menuWithItems:saveItem, resetItem, playLevelItem, nil];
    [menu alignItemsHorizontallyWithPadding:winSize.width*0.1f];
    menu.position = CGPointMake(winSize.width/2, saveItem.contentSize.height/2);
    [self addChild:menu z:100];
}

-(void)save {
    [fileHandler saveFile];
}

-(void)resetLevel {
    [pineapplesSpriteSheet removeAllChildrenWithCleanup:YES];
    [ropeSpriteSheet removeAllChildrenWithCleanup:YES];
    [ropes removeAllObjects];
    selectedObject = nil;
    [connectedRopes removeAllObjects];
    [fileHandler reset];
    [self drawLoadedLevel];
}

-(void) playLevel {
	[[CCDirector sharedDirector] replaceScene: [HelloWorldLayer sceneWithFileHandler: fileHandler]];
}

-(void)togglePopupMenu:(CGPoint)touchLocation {
    if (!popupMenu) {
        popupMenu = [[PopupMenu alloc] initWithParent:self];
        popupMenu.delegate = self;
    }
    if (popupMenu.isEnabled) {
        [popupMenu setMenuEnabled:NO];
    } else {
        [popupMenu setPopupPosition:touchLocation];
        [popupMenu setMenuEnabled:YES];
        if ([[pineapplesSpriteSheet children] count] < 1) {
            [popupMenu setRopeItemEnabled:NO];
        }
    }
}

-(void)createPineappleSpriteFromModel:(PineappleModel*) pineappleModel {
    CCSprite* pineappleSprite = [CCSprite spriteWithSpriteFrameName:@"pineapple.png"];
    pineappleSprite.tag = pineappleModel.id;
    CGPoint position = [CoordinateHelper levelPositionToScreenPosition:pineappleModel.position];
    pineappleSprite.position = position;
    [pineapplesSpriteSheet addChild:pineappleSprite];
}

-(void)createRopeSpriteFromModel:(RopeModel*)ropeModel {
    CGPoint anchorA;
    if (ropeModel.bodyAID == -1) {
        anchorA = [CoordinateHelper levelPositionToScreenPosition:ropeModel.anchorA];
    } else {
        PineappleModel* pineappleWithID = [fileHandler getPineappleWithID:ropeModel.bodyAID];
        anchorA = [CoordinateHelper levelPositionToScreenPosition:pineappleWithID.position];
    }
    
    CGPoint anchorB;
    if (ropeModel.bodyBID == -1) {
        anchorB = [CoordinateHelper levelPositionToScreenPosition:ropeModel.anchorB];
    } else {
        PineappleModel* pineappleWithID = [fileHandler getPineappleWithID:ropeModel.bodyBID];
        anchorB = [CoordinateHelper levelPositionToScreenPosition:pineappleWithID.position];
    }
    
    RopeSprite* ropeSprite = [[RopeSprite alloc] initWithParent:ropeSpriteSheet andRopeModel:ropeModel];
    [ropes addObject:ropeSprite];
}

-(void)setMode:(editorMode)newMode {
    mode = newMode;
    switch (mode) {
        case kRopeAnchorPineappleMode:
            background.color = kDefaultDisabledColor;
            for (CCSprite* pineapple in [pineapplesSpriteSheet children]) {
                pineapple.color = kDefaultSelectableColor;
            }
            break;
			
        case kRopeAnchorAnyMode:
            background.color = kDefaultSelectableColor;
            for (CCSprite* pineapple in [pineapplesSpriteSheet children]) {
                if (pineapple.tag == newRope.bodyAID) {
                    pineapple.color = kDefaultDisabledColor;
                } else {
                    pineapple.color = kDefaultSelectableColor;
                }
            }
            break;
			
        case kEditMode:
        default:
            background.color = kDefaultBackgroundColor;
            for (CCSprite* pineapple in [pineapplesSpriteSheet children]) {
                pineapple.color = ccc3(255, 255, 255);
            }
            break;
    }
}

-(CCSprite*)pineappleAtPosition:(CGPoint)position {
    for (CCSprite* pineapple in [pineapplesSpriteSheet children]) {
        if (CGRectContainsPoint(pineapple.boundingBox, position)) {
            return pineapple;
        }
    }
    return nil;
}

-(RopeSprite*)ropeAtPosition:(CGPoint)position {
    float minDistanceToRope = 5;
    RopeSprite* closestRope;
    for (RopeSprite* ropeSprite in ropes) {
        if (CGRectContainsPoint([ropeSprite getBoundingBox], position) && [ropeSprite distanceToPoint:position] < minDistanceToRope) {
            closestRope = ropeSprite;
            minDistanceToRope = [ropeSprite distanceToPoint:position];
        }
    }
    return closestRope;
}

-(void)removeRopeAtPosition:(CGPoint)position {
    RopeSprite* selectedRope = [self ropeAtPosition:position];
    if (selectedRope) {
        // remove rope (from ropes, ropesSpriteSheet, fileHandler
        [selectedRope cleanupSprite];
        [fileHandler removeRopeWithID:selectedRope.id];
        [ropes removeObject:selectedRope];
        return;
    }
}

-(void)removePineappleAtPosition:(CGPoint)position {
    CCSprite* selectedPineapple = [self pineappleAtPosition:position];
    if (selectedPineapple) {
        [fileHandler removePineappleWithID:selectedPineapple.tag];
        [selectedPineapple removeFromParentAndCleanup:YES];
        
        // remove all connected ropes
        NSArray* connectedRopesToBeRemoved = [self getAllRopesConnectedToPineappleWithID:selectedPineapple.tag];
        for (RopeSprite* rope in connectedRopesToBeRemoved) {
            [fileHandler removeRopeWithID:rope.id];
            [rope cleanupSprite];
            [ropes removeObject:rope];
        }
    }
}

-(NSMutableArray*)getAllRopesConnectedToPineappleWithID:(int)pineappleID {
    return [self getAllRopesConnectedToPineappleWithID:pineappleID setSelectedAnchor:NO];
}

-(NSMutableArray*)getAllRopesConnectedToPineappleWithID:(int) pineappleID setSelectedAnchor:(BOOL) setSelectedAnchor {
    NSMutableArray* tempRopes = [NSMutableArray arrayWithCapacity:5];
    for (RopeSprite* rope in ropes) {
        int bodyAID = [rope getBodyAID];
        int bodyBID = [rope getBodyBID];
        
        if (pineappleID == bodyAID || pineappleID == bodyBID) {
            [tempRopes addObject:rope];
        }
        if (setSelectedAnchor) {
            if (pineappleID == bodyAID) {
                [rope setSelectedAnchorType:kAnchorA];
            } else if (pineappleID == bodyBID) {
                [rope setSelectedAnchorType:kAnchorB];
            }
        }
    }
    return tempRopes;
}

-(void)selectFirstAnchor:(CGPoint)touchLocation {
    // if user tapped on pineapple set it as anchor of the new rope
    CCSprite* tappedPineapple = [self pineappleAtPosition:touchLocation];
    if (tappedPineapple) {
        [newRope setAnchorA:[CoordinateHelper screenPositionToLevelPosition:tappedPineapple.position]];
        newRope.bodyAID = tappedPineapple.tag;
        [self setMode:kRopeAnchorAnyMode];
    }
}

-(void)selectSecondAnchor:(CGPoint)touchLocation {
    // set second end of rope, can be either background or other pinapple, but not same pinapple as first one
    CCSprite* tappedPineapple = [self pineappleAtPosition:touchLocation];
    if (tappedPineapple && tappedPineapple.tag != newRope.bodyAID) {
        [newRope setAnchorB:[CoordinateHelper screenPositionToLevelPosition:tappedPineapple.position]];
        newRope.bodyBID = tappedPineapple.tag;
    }
    if (!tappedPineapple) {
        [newRope setAnchorB:[CoordinateHelper screenPositionToLevelPosition:touchLocation]];
        newRope.bodyBID = -1;
    }
    [self createRopeSpriteFromModel:newRope];
    [fileHandler addRopeFromModel: newRope];
    [self setMode:kEditMode];
}

-(void)selectRopeAnchor:(CGPoint)touchLocation {
    for (RopeSprite* rope in ropes) {
        float distanceToAnchorA = ccpDistance([rope getAnchorA], touchLocation);
        if (distanceToAnchorA < 10) {
            selectedObject = rope;
            [rope setSelectedAnchorType:kAnchorA];
            originalRopeAnchor = [rope getAnchorA];
        }
        float distanceToAnchorB = ccpDistance([rope getAnchorB], touchLocation);
        if (distanceToAnchorB < 10) {
            selectedObject = rope;
            [rope setSelectedAnchorType:kAnchorB];
            originalRopeAnchor = [rope getAnchorB];
        }
        if (selectedObject) {
            break;
        }
    }
}

-(void)selectPineapple:(CGPoint)touchLocation {
    CCSprite* sprite = [self pineappleAtPosition:touchLocation];
    if (sprite) {
        selectedObject = sprite;
        // remember which ropes are tied to the selectedSprite
        connectedRopes = [self getAllRopesConnectedToPineappleWithID:sprite.tag setSelectedAnchor:YES];
    }
}

#pragma mark - Touch Handling
-(void)longPress:(UILongPressGestureRecognizer*)longPressGestureRecognizer {
    if (mode == kEditMode && longPressGestureRecognizer.state == UIGestureRecognizerStateBegan) {
        NSLog(@"longpress began");
        NSValue* locationObject = [longPressGestureRecognizer valueForKey:@"_startPointScreen"];
        CGPoint touchLocation = [locationObject CGPointValue];
        touchLocation = [[CCDirector sharedDirector] convertToGL:touchLocation];
        [self removeRopeAtPosition:touchLocation];
        [self removePineappleAtPosition:touchLocation];
    }
}

-(BOOL)ccTouchBegan:(UITouch *)touch withEvent:(UIEvent *)event {
    CGPoint touchLocation = [touch locationInView:touch.view];
    touchLocation = [[CCDirector sharedDirector] convertToGL:touchLocation];
    
    switch (mode) {
        case kEditMode:
            // check if a rope's anchor point was tapped, if yes select it
            [self selectRopeAnchor:touchLocation];
            if (!selectedObject) {
                [self selectPineapple:touchLocation];
            }
            if (!selectedObject && ![self ropeAtPosition:touchLocation]) {
                [self togglePopupMenu:touchLocation];
            }
            break;
        case kRopeAnchorPineappleMode:
            [self selectFirstAnchor:touchLocation];
            break;
        case kRopeAnchorAnyMode:
            [self selectSecondAnchor:touchLocation];
            break;
    }
    return YES;
}

-(void)ccTouchMoved:(UITouch *)touch withEvent:(UIEvent *)event {
    if (selectedObject && mode == kEditMode) {
        CGPoint touchLocation = [touch locationInView:touch.view];
        touchLocation = [[CCDirector sharedDirector] convertToGL:touchLocation];
        
        CGPoint oldTouchLocation = [touch previousLocationInView:touch.view];
        oldTouchLocation = [[CCDirector sharedDirector] convertToGL:oldTouchLocation];
        
        if ([selectedObject isMemberOfClass:RopeSprite.class]) {
            RopeSprite* rope = (RopeSprite*) selectedObject;
            [rope moveSelectedAnchorTo:touchLocation];
        }
		
        CGPoint movementVector = ccpSub(touchLocation, oldTouchLocation);
        if ([selectedObject isMemberOfClass:CCSprite.class]) {
            CCSprite* selectedSprite = (CCSprite*) selectedObject;
            CGPoint newPosition = ccpAdd(selectedSprite.position, movementVector);
            selectedSprite.position = newPosition;
            // remember which ropes are tied to the selectedSprite
            for (RopeSprite* rope in connectedRopes) {
                [rope moveSelectedAnchorTo:newPosition];
            }
        }
    }
}

-(void)ccTouchEnded:(UITouch *)touch withEvent:(UIEvent *)event {
    if (mode == kEditMode) {
        if ([selectedObject isMemberOfClass: RopeSprite.class]) {
            RopeSprite* rope = (RopeSprite*) selectedObject;
            CGPoint touchLocation = [touch locationInView:touch.view];
            touchLocation = [[CCDirector sharedDirector] convertToGL:touchLocation];
            
            // snap to pineapple if rope ends up on top of it
            int newAnchorID = -1;
            CCSprite* pineappleAtTouchLocation = [self pineappleAtPosition:touchLocation];
            if (pineappleAtTouchLocation) {
                [rope moveSelectedAnchorTo:pineappleAtTouchLocation.position];
                // get id of pineapple to properly set anchor in connected data model
                newAnchorID = pineappleAtTouchLocation.tag;
            }
            
            BOOL isValidNewAnchor = [rope isValidNewAnchorID:newAnchorID];
            if (isValidNewAnchor) {
                [fileHandler moveRopeWithId:rope.id withAnchorType:[rope getSelectedAnchorType] withAnchorId:newAnchorID to:[CoordinateHelper screenPositionToLevelPosition:touchLocation]];
            } else {
                [fileHandler moveRopeWithId:rope.id withAnchorType:[rope getSelectedAnchorType] withAnchorId:[rope getSelectedAnchorType] to:[CoordinateHelper screenPositionToLevelPosition:originalRopeAnchor]];
                [rope moveSelectedAnchorTo:originalRopeAnchor];
            }
            
            // set selected anchor to none
            [rope setSelectedAnchorType:kAnchorNone];
        }
        if ([selectedObject isMemberOfClass: CCSprite.class]) {
            CCSprite* pineappleSprite = (CCSprite*) selectedObject;
            int pineappleID = pineappleSprite.tag;
            CGPoint newPineapplePosition = [CoordinateHelper screenPositionToLevelPosition:pineappleSprite.position];
            [fileHandler movePineappleWithId:pineappleID to:newPineapplePosition];
        }
        [connectedRopes removeAllObjects];
        selectedObject = nil;
    }
}

#pragma mark - PopupMenu Delegates
-(void)createPineappleAt:(CGPoint)position {
    CGSize winSize = [CCDirector sharedDirector].winSize;
    PineappleModel* pineappleModel = [fileHandler addPineappleAt:CGPointMake(position.x/winSize.width, position.y/winSize.height)];
    [self createPineappleSpriteFromModel:pineappleModel];
    [popupMenu setMenuEnabled:NO];
}

-(void)createRopeAt:(CGPoint)position {
    [self setMode:kRopeAnchorPineappleMode];
    newRope = [[RopeModel alloc] init];
    [popupMenu setMenuEnabled:NO];
}

@end
