//
//  LevelFileHandler.m
//  CutTheVerlet
//
//  Created by Fahim Farook on 23/3/13.
//  Copyright (c) 2013 OpenTheJob. All rights reserved.
//

#import "LevelFileHandler.h"
#import "PineappleModel.h"
#import "RopeModel.h"
#import "FileHelper.h"
#import "GDataXMLNode.h"

@interface LevelFileHandler () {
    NSString* _filename;
}

@end

@implementation LevelFileHandler

+(void)sortArrayById:(NSMutableArray*)objectArray {
    NSSortDescriptor* sortDescriptor = [[NSSortDescriptor alloc] initWithKey:@"id" ascending:YES];
    NSArray* sortDescriptors = [NSArray arrayWithObject:sortDescriptor];
    [objectArray sortUsingDescriptors:sortDescriptors];
}

+(NSNumber*)firstUnusedIdInArray:(NSArray*)array {
    NSNumber* firstUnusedID = nil;
    int lastID = 0;
    for (AbstractModel* object in array) {
        if (object.id - lastID > 1) {
            firstUnusedID = [NSNumber numberWithInt:lastID + 1];
            break;
        }
        lastID++;
    }
    if (!firstUnusedID) {
        firstUnusedID = [NSNumber numberWithInt:lastID + 1];
    }
    return firstUnusedID;
}

-(id)initWithFileName:(NSString*)filename {
    self = [super init];
    if (self) {
		_filename = filename;
        [self loadFile];
    }
    return self;
}

/* loads an XML file containing level data */
-(void) loadFile {
    // load file from documents directory if possible, if not try to load from mainbundle
    NSString *filePath = [FileHelper dataFilePathForFileWithName:_filename withExtension:@".xml" forSave:NO];
    NSData *xmlData = [[NSMutableData alloc] initWithContentsOfFile:filePath];
    GDataXMLDocument *doc = [[GDataXMLDocument alloc] initWithData:xmlData options:0 error:nil];
	
	// clean level data before loading level from file
    self.pineapples = [NSMutableArray arrayWithCapacity:5];
    self.ropes = [NSMutableArray arrayWithCapacity:5];
	
    // if there is no file doc will be empty and we simply return from this method
    if (doc == nil) {
		return;
	}
    NSLog(@"%@", doc.rootElement);
	
    NSArray* pineappleElements = [doc.rootElement elementsForName:@"pineapple"];
    for (GDataXMLElement* pineappleElement in pineappleElements) {
        PineappleModel* pineappleModel = [[PineappleModel alloc] init];
        // load id
        pineappleModel.id = [pineappleElement attributeForName:@"id"].stringValue.intValue;
        // load level coordinate, for display on screen needs to be multiplied with screen size
        float x = [pineappleElement attributeForName:@"x"].stringValue.floatValue;
        float y = [pineappleElement attributeForName:@"y"].stringValue.floatValue;
        pineappleModel.position = CGPointMake(x, y);
        // load damping if set, otherwise keep default value
        GDataXMLNode* dampingElement = [pineappleElement attributeForName:@"damping"];
        if (dampingElement) {
            pineappleModel.damping = [pineappleElement attributeForName:@"damping"].stringValue.floatValue;
        }
        [self.pineapples addObject:pineappleModel];
    }
	
	NSArray* ropesElement = [doc.rootElement elementsForName:@"rope"];
    // ids for ropes start at 1 and are given out in the file handler
    // they are not stored in the XML file as they are only needed for the editor
    // and do not convey any substantial information about the level layout
    int ropeID = 1;
    for (GDataXMLElement* ropeElement in ropesElement) {
        RopeModel* ropeModel = [[RopeModel alloc] init];
        ropeModel.id = ropeID;
        // load the anchor points consisting of the body id the rope is tied to
        // (-1 stands for the background) and the position, which will be ignored by the game
        // later on if the rope is tied to a pineapple
        GDataXMLElement* anchorA = [[ropeElement elementsForName:@"anchorA"] objectAtIndex:0];
        ropeModel.bodyAID = [anchorA attributeForName:@"body"].stringValue.intValue;
        
        float ax;
        float ay;
        if (ropeModel.bodyAID == -1) {
            ax = [anchorA attributeForName:@"x"].stringValue.floatValue;
            ay = [anchorA attributeForName:@"y"].stringValue.floatValue;
        } else {
            PineappleModel* pineappleModel = [self getPineappleWithID:ropeModel.bodyAID];
            ax = pineappleModel.position.x;
            ay = pineappleModel.position.y;
        }
        
        //[LevelFileHandler getFloatValueForName:@"y" fromElement:anchorA];
        ropeModel.anchorA = CGPointMake(ax, ay);
        
        GDataXMLElement* anchorB = [[ropeElement elementsForName:@"anchorB"] objectAtIndex:0];
        ropeModel.bodyBID = [anchorB attributeForName:@"body"].stringValue.intValue;
        
        float bx;
        float by;
        if (ropeModel.bodyBID == -1) {
            bx = [anchorB attributeForName:@"x"].stringValue.floatValue;
            by = [anchorB attributeForName:@"y"].stringValue.floatValue;
        } else {
            PineappleModel* pineappleModel = [self getPineappleWithID:ropeModel.bodyBID];
            bx = pineappleModel.position.x;
            by = pineappleModel.position.y;
        }
        
        ropeModel.anchorB = CGPointMake(bx, by);
        
        GDataXMLNode* sagityElement = [ropeElement attributeForName:@"sagity"];
        if (sagityElement) {
            ropeModel.sagity = [ropeElement attributeForName:@"sagity"].stringValue.floatValue;
        }
        
        [self.ropes addObject:ropeModel];
        
        // increase ropeID as the ids need to be unique
        ropeID++;
    }
}

+(AbstractModel*) getModelWithID:(int) id fromArray:(NSArray*) array {
    for (AbstractModel* model in array) {
        if (model.id == id) {
            return model;
        }
    }
    return nil;
}

-(PineappleModel*) getPineappleWithID:(int)id {
    return (PineappleModel*)[LevelFileHandler getModelWithID:id fromArray:self.pineapples];
}

-(PineappleModel*)addPineappleAt:(CGPoint)position {
    NSNumber* firstUnusedID = [LevelFileHandler firstUnusedIdInArray:self.pineapples];
    PineappleModel* newPineapple = [[PineappleModel alloc] init];
    newPineapple.id = [firstUnusedID intValue];
    newPineapple.position = position;
    [self.pineapples addObject:newPineapple];
    return newPineapple;
}

-(void)addRopeFromModel:(RopeModel*)newRope {
    [LevelFileHandler sortArrayById: self.ropes];
    if (!newRope.id) {
        NSNumber* firstUnusedID = [LevelFileHandler firstUnusedIdInArray:self.ropes];
        newRope.id = firstUnusedID.intValue;
    }
    [self.ropes addObject:newRope];
}

-(void)removeRopeWithID:(int)id {
    RopeModel* ropeToBeRemoved;
    for (RopeModel* rope in self.ropes) {
        if (rope.id == id) {
            ropeToBeRemoved = rope;
            break;
        }
    }
    [self.ropes removeObject:ropeToBeRemoved];
}

-(void)removePineappleWithID:(int)id {
    PineappleModel* pineappleToBeRemoved;
    for (PineappleModel* pineapple in self.pineapples) {
        if (pineapple.id == id) {
            pineappleToBeRemoved = pineapple;
            break;
        }
    }
    [self.pineapples removeObject:pineappleToBeRemoved];
}

-(void)moveRopeWithId:(int)id withAnchorType:(anchorType)anchor withAnchorId:(int)anchorId to:(CGPoint)position {
    for (RopeModel* ropeModel in self.ropes) {
        if (ropeModel.id == id) {
            switch (anchor) {
                case kAnchorA:
                    ropeModel.anchorA = position;
                    ropeModel.bodyAID = anchorId;
                    break;
                case kAnchorB:
                    ropeModel.anchorB = position;
                    ropeModel.bodyBID = anchorId;
                    break;
                default:
                    break;
            }
        }
    }
}

-(void)movePineappleWithId:(int)id to:(CGPoint)position {
    for (PineappleModel* pineappleModel in self.pineapples) {
        if (pineappleModel.id == id) {
            pineappleModel.position = position;
            break;
        }
    }
}

-(void) saveFile {
    // create empty xml structure
    GDataXMLElement* levelElement = [GDataXMLElement elementWithName:@"level"];
    // add all Pineapple Elements
    for (PineappleModel* pineapple in self.pineapples) {
        GDataXMLElement* pineappleElement = [GDataXMLElement elementWithName:@"pineapple"];
        GDataXMLElement* idAttribute = [GDataXMLElement elementWithName:@"id" stringValue:[NSString stringWithFormat:@"%i", pineapple.id]];
        GDataXMLElement* xAttribute = [GDataXMLElement elementWithName:@"x" stringValue:[NSString stringWithFormat:@"%.3f", pineapple.position.x]];
        GDataXMLElement* yAttribute = [GDataXMLElement elementWithName:@"y" stringValue:[NSString stringWithFormat:@"%.3f", pineapple.position.y]];
        if (pineapple.damping != kDefaultDamping) {
            GDataXMLElement* dampingAttribute = [GDataXMLElement elementWithName:@"damping" stringValue:[NSString stringWithFormat:@"%.1f", pineapple.damping]];
            [pineappleElement addAttribute:dampingAttribute];
        }
        [pineappleElement addAttribute:idAttribute];
        [pineappleElement addAttribute:xAttribute];
        [pineappleElement addAttribute:yAttribute];
        [levelElement addChild:pineappleElement];
    }
    // Add all rope elements
    for (RopeModel* rope in self.ropes) {
        GDataXMLElement* ropeElement = [GDataXMLElement elementWithName:@"rope"];
        
        //Anchor A
        GDataXMLElement* anchorAElement = [GDataXMLElement elementWithName:@"anchorA"];
        GDataXMLElement* bodyAttributeA = [GDataXMLElement elementWithName:@"body" stringValue:[NSString stringWithFormat:@"%i", rope.bodyAID]];
        [anchorAElement addAttribute:bodyAttributeA];
        if (rope.bodyAID == -1) {
            GDataXMLElement* xAttributeA = [GDataXMLElement elementWithName:@"x" stringValue:[NSString stringWithFormat:@"%.3f", rope.anchorA.x]];
            GDataXMLElement* yAttributeA = [GDataXMLElement elementWithName:@"y" stringValue:[NSString stringWithFormat:@"%.3f", rope.anchorA.y]];
            [anchorAElement addAttribute:xAttributeA];
            [anchorAElement addAttribute:yAttributeA];
        }
        [ropeElement addChild:anchorAElement];
        
		// Anchor B
        GDataXMLElement* anchorBElement = [GDataXMLElement elementWithName:@"anchorB"];
        GDataXMLElement* bodyAttributeB = [GDataXMLElement elementWithName:@"body" stringValue:[NSString stringWithFormat:@"%i", rope.bodyBID]];
        [anchorBElement addAttribute:bodyAttributeB];
        if (rope.bodyBID == -1) {
            GDataXMLElement* xAttributeB = [GDataXMLElement elementWithName:@"x" stringValue:[NSString stringWithFormat:@"%.3f", rope.anchorB.x]];
            GDataXMLElement* yAttributeB = [GDataXMLElement elementWithName:@"y" stringValue:[NSString stringWithFormat:@"%.3f", rope.anchorB.y]];
            [anchorBElement addAttribute:xAttributeB];
            [anchorBElement addAttribute:yAttributeB];
        }
        [ropeElement addChild:anchorBElement];
		
        [levelElement addChild:ropeElement];
    }
    // Write to file in document directory
    GDataXMLDocument *document = [[GDataXMLDocument alloc] initWithRootElement:levelElement];
    NSData *xmlData = document.XMLData;
    
    // creates an empty folder where all the levels should be stored, if necessary
    [FileHelper createFolder:@"levels"];
    NSString* filePath = [FileHelper dataFilePathForFileWithName:_filename withExtension:@".xml" forSave:YES];
    [xmlData writeToFile:filePath atomically:YES];
}

-(void)reset {
    [self loadFile];
}

@end
